/* LICENSE BLOCK
Part of mde: a Modular D game-oriented Engine
Copyright © 2007-2008 Diggory Hardy

This program is free software: you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation, either
version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/// Gui layout widgets.
module mde.gui.widget.layout;

import mde.gui.widget.Widget;
import mde.gui.exception;

import tango.io.Stdout;
debug {
    import tango.util.log.Log : Log, Logger;
    private Logger logger;
    static this () {
        logger = Log.getLogger ("mde.gui.widget.layout");
    }
}

/** Encapsulates a grid of Widgets.
*
* Currently there is no support for changing number of cells, sub-widgets or sub-widget properties
* (namely isW/HSizable and minimal size) after this() has run.
*
* Since a grid with either dimension zero is not useful, there must be at least one sub-widget.
*
* The grid has no border but has spacing between widgets. */
class GridLayoutWidget : Widget
{
    this (IWindow wind, int[] data) {
        // Get grid size and check data
        // Check sufficient data for rows, cols, and at least one widget:
        if (data.length < 4) throw new WidgetDataException;
        super (wind, data);
        
        rows = data[1];
        cols = data[2];
        if (data.length != 3 + rows * cols) throw new WidgetDataException;
        /* data.length >= 3 so besides checking the length is correct, this tells us:
         *      rows * cols >= 4 - 3 = 1            a free check!
         * The only thing not checked is whether both rows and cols are negative, which would
         * cause an exception when dynamic arrays are allocated later (and is unlikely). */
        
        // Get all sub-widgets
        subWidgets.length = rows*cols;
        foreach (i, ref subWidget; subWidgets) {
            subWidget = window.makeWidget (data[i+3]);
        }
        
        // Calculate cached construction data
        genCachedConstructionData;
    }
    
    int[] adjust (int[] data) {
        // Give all sub-widgets their data:
        foreach (widget; subWidgets)
            data = widget.adjust (data);
        
        /* We basically short-cut setSize by loading previous col/row sizes and doing the final
         * calculations.
         * Note: if setSize gets called afterwards, it should have same dimensions and so not do
         * anything. */
        
        int lenUsed = 0;
        if (data.length < rows + cols) {    // data error; use defaults
            colW = colWMin.dup;
            rowH = rowHMin.dup;
        } else {                            // sufficient data
            lenUsed = rows+cols;
            colW = data[0..cols];
            rowH = data[cols..lenUsed];
            
            // Check row sizes are valid:
            //NOTE: this could be made optional
            //NOTE: could also check non-resizable sizes are not too large
            foreach (i, ref w; colW)
                if (w < colWMin[i]) w = colWMin[i];
            foreach (i, ref h; rowH)
                if (h < rowHMin[i]) h = rowHMin[i];
        }
        
        genCachedMutableData;
        w = colW[$-1] + colX[$-1];
        h = rowY[$-1] + rowH[$-1];
        
        return data[lenUsed..$];
    }
    
    int[] getCreationData () {
        int[] ret;
        ret.length = 3 + subWidgets.length;
        
        ret [0..3] = [widgetType, rows, cols];  // first data
        
        foreach (i,widget; subWidgets)          // sub widgets
            ret[i+3] = window.addCreationData (widget);
        
        return ret;
    }
    int[] getMutableData () {
        int[] ret;
        foreach (widget; subWidgets)
            ret ~= widget.getMutableData;
        
        ret ~= colW ~ rowH;
        return ret;
    }
    
    bool isWSizable () {
        return (sizableCols.length != 0);
    }
    
    bool isHSizable () {
        return (sizableRows.length != 0);
    }
    
    /* Calculates the minimal size from all rows and columns of widgets. */
    void getMinimalSize (out int mw, out int mh) {
        mw = this.mw;
        mh = this.mh;
    }
    
    void setSize (int nw, int nh) {
        // Step 1: calculate the row/column sizes.
        w += adjustCellSizes (colW, colWMin, sizableCols, nw - w, true);
        h += adjustCellSizes (rowH, rowHMin, sizableRows, nh - h, true);
        
        // Step 2: calculate the row/column offsets (positions) and set the sub-widget's sizes.
        genCachedMutableData;
        
        // Step 3: position needs to be set
        // Currently this happens by specifying that setPosition should be run after setSize.
    }
    
    void setPosition (int x, int y) {
        this.x = x;
        this.y = y;
        
        foreach (i,widget; subWidgets)
            widget.setPosition (x + colX[i % cols], y + rowY[i / cols]);
    }
    
    
    // Find the relevant widget.
    IWidget getWidget (int cx, int cy) {
        int lx = cx - x, ly = cy - y;       // use coords relative to this widget
        
        // Find the column
        int i = cols - 1;                   // starting from right...
        while (lx < colX[i]) {              // decrement while left of this column
            debug assert (i > 0, "getWidget: left of first column");  // should be impossible
            --i;
        }                                   // now (lx >= colX[i])
        if (lx >= colX[i] + colW[i]) return this;   // between columns
        
        // Find the row;
        int j = rows - 1;
        while (ly < rowY[j]) {
            debug assert (j > 0, "getWidget: above first row");   // should be impossible
            --j;
        }
        if (ly >= rowY[j] + rowH[j]) return this;
        
        // Now we know it's in widget (i,j)'s cell (but the widget may not take up the whole cell)
        lx -= colX[i];
        ly -= rowY[j];
        IWidget widg = subWidgets[i + j*cols];
        widg.getCurrentSize (i,j);
        if (lx < i && ly < j)
            return widg.getWidget (cx, cy);
        return this;    // wasn't in cell
    }
    
    // Resizing columns & rows
    void clickEvent (ushort cx, ushort cy, ubyte b, bool state) {
        debug scope (failure)
                logger.warn ("clickEvent: failure");
        if (b == 1 && state == true) {
            /* Note: Because of getWidget, this function is only called if the click is not on a
            * sub-widget, so we know it's on some divisor (so at least one of resizeCol and
            * resizeRow is non-negative). */
            
            // Find the column
            if (sizableCols.length != 0) {
                int l = cx - x;                 // use relative coords
                size_t i = cols - 1;            // index, from right
                while (l < colX[i]) {           // decrement while left of this column
                    debug assert (i > 0, "clickEvent: left of first column");
                    --i;
                }                               // now (l >= colX[resizeCol])
                if (l < colX[i] + colW[i]) i = -1;  // on a sub-widget
                
                // Set resizeColsL / resizeColsH
                // Want to find j such that [0..j],[j..$] divide sizableCols about i:
                size_t j = 0;
                while (j < sizableCols.length && sizableCols[j] <= i) ++j;
                
                resizeColsL = sizableCols[0..j];
                resizeColsH = sizableCols[j..$];
                
                // Cannot resize if either list is empty. resizeCallback checks the length of L,
                // but to save it checking R too, we set L's length zero if R's is.
                if (resizeColsH.length == 0)
                    resizeColsL = null;
            }
            
            // Find the row
            if (sizableRows.length != 0) {
                int l = cy - y;
                size_t i = rows - 1;
                while (l < rowY[i]) {
                    debug assert (i > 0, "clickEvent: above first row");
                    --i;
                }
                if (l < rowY[i] + rowH[i]) i = -1;
                
                size_t j = 0;
                while (j < sizableRows.length && sizableRows[j] <= i) ++j;
                
                resizeRowsL = sizableRows[0..j];
                resizeRowsH = sizableRows[j..$];
                
                if (resizeRowsH.length == 0)
                    resizeRowsL = null;
            }
            
            if (resizeColsL is null && resizeRowsL is null)
                return;     // no resizing to do
            
            dragX = cx;
            dragY = cy;
            
            window.gui.addClickCallback (&endCallback);
            window.gui.addMotionCallback (&resizeCallback);
        }
    }
    
    void draw () {
        super.draw ();
        
        foreach (widget; subWidgets)
            widget.draw ();
    }
    
private:
    //BEGIN Cache calculation functions
    /* Calculations which need to be run whenever a new sub-widget structure is set
     * (i.e. to produce cached data calculated from construction data). */
    void genCachedConstructionData () {
        // Calculate the minimal column and row sizes:
        colWMin = new int[cols];    // set length, making sure the arrays are initialised to zero
        rowHMin = new int[rows];
        int ww, wh;     // sub-widget minimal sizes
        foreach (i,widget; subWidgets) {
            widget.getMinimalSize (ww, wh);
            
            // Increase dimensions if current minimal size is larger:
            uint n = i % cols;      // column
            if (colWMin[n] < ww) colWMin[n] = ww;
            n = i / cols;           // row
            if (rowHMin[n] < wh) rowHMin[n] = wh;
        }
        
        
        // Calculate the overall minimal size, starting with the spacing:
        mh = window.renderer.layoutSpacing;     // use mh temporarily
        mw = mh * (cols - 1);
        mh *= (rows - 1);
        
        foreach (x; colWMin)            // add the column/row's dimensions
            mw += x;
        foreach (x; rowHMin)
            mh += x;
        
        
        // Find which cols/rows are resizable:
        sizableCols = sizableRows = null;   // reset; we're about to concatenate to them
        
        forCols:
        for (uint i = 0; i < cols; ++i) {                       // for each column
            for (uint j = 0; j < subWidgets.length; j += cols)  // for each row
                if (!subWidgets[i+j].isWSizable)        // column not resizable
                    continue forCols;                   // continue the outer for loop
                
            // column is resizable if we get to here
            sizableCols ~= i;
        }
        
        forRows:
        for (uint i = 0; i < subWidgets.length; i += cols) {    // for each row
            for (uint j = 0; j < cols; ++j)                     // for each column
                if (!subWidgets[i+j].isHSizable)
                    continue forRows;
            
            sizableRows ~= i / cols;
        }
    }
    
    /* Calculations which need to be run whenever resizing occurs (or deeper alterations)
     * (i.e. to produce cached data calculated from construction and mutable data). */
    void genCachedMutableData () {
        // Calculate column and row locations:
        colX.length = cols;
        rowY.length = rows;
        int spacing = window.renderer.layoutSpacing;
        
        int cum = 0;
        foreach (i, x; rowH) {
            rowY[i] = cum;
            cum += x + spacing;
        }
        
        cum = 0;
        foreach (i, x; colW) {
            colX[i] = cum;
            cum += x + spacing;
        }
        
        // Tell subwidgets their new sizes:
        foreach (i,widget; subWidgets)
            widget.setSize (colW[i % cols], rowH[i / cols]);
    }
    //END Cache calculation functions
    
    /* Adjust the total size of rows/columns (including spacing) by diff.
     *
     * Params:
     *  cellD       = current sizes; is adjusted by the function to new sizes
     *  cellDMin    = minimal sizes (see colWMin/rowHMin)
     *  sizableCells= List of indexes in cellD for cells which are resizable
     *  diff        = amount to increase/decrease the total size
     *  startHigh   = if true, start resizing from the tail end of cellD, instead of the beginning
     *
     * Returns:
     *  The amount adjusted. This may be larger than diff, since cellD is clamped by cellDMin.
     */
    int adjustCellSizes (ref int[] cellD, ref int[] cellDMin, ref size_t[] sizableCells, int diff, bool startHigh)
    in {// Could occur if adjust isn't called first, but this would be a code error:
        assert (cellD !is null, "adjustCellSizes: cellD is null");
    } body {
        // Cannot resize if no cells are sizable:
        if (sizableCells.length == 0) return 0;
        
        size_t si = (startHigh ? sizableCells.length-1 : 0);    // starting index of sizableCells
        
        // FIXME: could do with an overhaul
        if (diff > 0) {         // increase size of first resizable cell
            cellD[sizableCells[si]] += diff;
            return diff;
        }
        else if (diff < 0) {    // decrease
            size_t sc = (startHigh ? -1 : 1);   // amount to iterate
            size_t ci;          // index in cellD
            int rd = diff;      // running diff
            while (true) {
                ci = sizableCells[si];
                
                cellD[ci] += rd; // decrease this cell's size (but may be too much)
                rd = cellD[ci] - cellDMin[ci];
                if (rd >= 0)    // OK; we're done
                    return diff;// we hit the mark exactly
                
                // else we decreased it too much!
                cellD[ci] = cellDMin[ci];
                // rd is remainder to decrease by
                
                si += sc;       // iterate
                if (si < 0 || si >= sizableCells.length)        // run out of next cells
                    return diff - rd;   // still had rd left to decrease
            }
        }
        else                   // no adjustment needed
            return 0;
    }
    
    //BEGIN Col/row resizing
    void resizeCallback (ushort cx, ushort cy) {
        int move;           // NOTE: all resizing is relative, unlike in Window
        
        if (resizeColsL.length != 0) {
            move = cx - dragX;
            
            // do shrinking first (in case we hit the minimum)
            // Note: we rely on x[a..b] pointing to the same memory as x
            if (move < 0) {
                move = -adjustCellSizes (colW, colWMin, resizeColsL, move, false);
                adjustCellSizes (colW, colWMin, resizeColsH, move, true);
            } else {
                move = -adjustCellSizes (colW, colWMin, resizeColsH, -move, true);
                adjustCellSizes (colW, colWMin, resizeColsL, move, false);
            }
            
            dragX = cx;
        }
        
        if (resizeRowsL.length != 0) {
            move = cy - dragY;
            
            // do shrinking first (in case we hit the minimum)
            // Note: we rely on x[a..b] pointing to the same memory as x
            if (move < 0) {
                move = -adjustCellSizes (rowH, rowHMin, resizeRowsL, move, false);
                adjustCellSizes (rowH, rowHMin, resizeRowsH, move, true);
            } else {
                move = -adjustCellSizes (rowH, rowHMin, resizeRowsH, -move, true);
                adjustCellSizes (rowH, rowHMin, resizeRowsL, move, false);
            }
            
            dragY = cy;
        }
        
        // NOTE: this might be able to be made more efficient, but basically this all needs to happen:
        genCachedMutableData;
        setPosition (x,y);
        window.requestRedraw;
    }
    bool endCallback (ushort cx, ushort cy, ubyte b, bool state) {
        if (b == 1 && state == false) {
            window.gui.removeCallbacks (cast(void*) this);
            // Remove unwanted data (although it shouldn't free any memory):
            resizeColsL = resizeColsH = resizeRowsL = resizeRowsH = null;
            return true;    // we've handled the up-click
        }
        return false;       // we haven't handled it
    }
    
protected:
    // Data for resizing cols/rows:
    int dragX, dragY;   // coords where drag starts
    // Lists of columns/rows with lower/higher index than the resize position
    size_t[] resizeColsL, resizeColsH, resizeRowsL, resizeRowsH;
    //END Col/row resizing
    
    
    //BEGIN Construction (i.e. non-mutable) data
    int cols, rows;     // number of cells in grid
    
    /* All widgets in the grid, by row. Order:  [ 0 1 ]
     *                                          [ 2 3 ] */
    IWidget[] subWidgets;
    
    // Cached data calculated from construction data:
    // Minimal column width / row height:
    int[] colWMin, rowHMin;
    int mw, mh;         // minimal dimensions
    
    // All resizable cols/rows, in order:
    size_t[] sizableCols, sizableRows;
    //END Construction data
    
    
    //BEGIN Mutable data
    int[] colW, rowH;   // column width / row height (largest size in col/row)
    
    // Cached data calculated from construction and mutable data:
    int[] colX, rowY;   // cumulative colW[i-1] + padding (add x to get column's left x-coord)
    //END Mutable data
}
